import datetime
from dateutil.parser import parse as dateparse


class BaseSetting:
    def __init_subclass__(cls, param_name, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._param_name = param_name

    @property
    def _alias(self):
        param = getattr(self, self.__class__._param_name)
        return param

    def __repr__(self):
        if self._alias is not None:
            return f'''[{self._param_name}:{self._fmt_param()}]'''
        return ""

    def _fmt_param(self):
        return self._alias



class PayloadFormat(BaseSetting, param_name="out"):
    def __init__(self, out):
        self.out = out


class Timeout(BaseSetting, param_name="timeout"):
    def __init__(self, timeout):
        self.timeout = timeout


class Maxsize(BaseSetting, param_name="maxsize"):
    def __init__(self, maxsize):
        self.maxsize = maxsize


class Bbox(BaseSetting, param_name="bbox"):
    def __init__(self, bbox):
        self.bbox = bbox

    def _fmt_param(self):
        s, w, n, e = self.bbox
        return f'''{s:.8f},{w:.8f},{n:.8f},{e:.8f}'''


class Date(BaseSetting, param_name="date"):
    def __init__(self, date):
        self.date = date

    def _fmt_param(self):
        return self.date.isoformat(timespec="seconds") + "Z"



class QuerySettings:
    DEFAULT_FORMAT = "xml"
    DEFAULT_TIMEOUT = 180
    MAXSIZE_LIMIT = 2 * 1073741824 # 2 GB is max size request

    def __init__(self,
                 payload_format=None,
                 timeout=None,
                 maxsize=None,
                 date=None,
                 bbox=None):

        self.payload_format = payload_format
        self.timeout = timeout
        self.maxsize = maxsize
        self.date = date
        self.bbox = bbox

    def __repr__(self):
        s = (
            f'''{self._out}'''
            f'''{self._timeout}'''
            f'''{self._maxsize}'''
            f'''{self._date}'''
            f'''{self._bbox}'''
        )
        if s != "":
            s += ";"
        return s

    @property
    def payload_format(self):
        return self._out.out

    @payload_format.setter
    def payload_format(self, val):
        self._out = PayloadFormat(val)

    @property
    def timeout(self):
        return self._timeout.timeout

    @timeout.setter
    def timeout(self, val):
        self._timeout = Timeout(val)

    @property
    def maxsize(self):
        return self._maxsize.maxsize

    @maxsize.setter
    def maxsize(self, val):
        self._maxsize = Maxsize(val)

    @property
    def date(self):
        return self._date.date

    @date.setter
    def date(self, val):
        self._date = Date(val)

    @property
    def bbox(self):
        return self._bbox.bbox

    @bbox.setter
    def bbox(self, val):
        self._bbox = Bbox(val)





import itertools
from .filters import CompoundFilter


class OSMElement:
    geom_type = NotImplemented

    @classmethod
    def _fmt(cls):
        return cls.__name__.lower()


class Node(OSMElement):
    _level = 0


class Way(OSMElement):
    _level = 1


class Relation(OSMElement):
    _level = 2





class AbstractQueryStatement:
    _input = None
    _output = None
    _default_name = "_"

    def __init__(self, inputs=None, name=None, filters=None):
        if inputs:
            self._inputs = inputs
        if name:
            self._name = name
        self.filters = CompoundFilter(filters)

    def _fmt_statement(self):
        raise NotImplementedError

    def __repr__(self):
        return f'''{self._fmt_statement()};'''

    def __add__(self, other):
        if isinstance(other, AbstractQueryStatement):
            return UnionQuerySet([self, other])
        raise NotImplementedError

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, AbstractQueryStatement):
            # Check output type match
            return DifferenceQuerySet([self, other])
        raise NotImplementedError

    def __rsub__(self, other):
        return NotImplemented

    def __isub__(self, other):
        return self.__sub__(other)

    def union(self, other):
        return self + other

    def difference(self, other):
        return self - other

    def iunion(self, other):
        self += other

    def idifference(self, other):
        self -= other



class GenericElementQuery(AbstractQueryStatement):
    _element = NotImplemented
    _input = None

    def __init__(self, name=None, filters=None):
        super().__init__(name=name, filters=filters)

    def add_filter(self, _filter):
        self.filters.add_filter(_filter)

    @property
    def _output(self):
        return self._element

    def _fmt_element(self):
        return f'''{self._element._fmt()}'''

    def _fmt_statement(self):
        msg = f'''{self._fmt_element()}'''
        if self.filters.filters:
            msg = f'''{msg}{self.filters}'''
        return msg


class CompoundQuery(GenericElementQuery):
    _elements = ()

    @property
    def _output(self):
        return self._elements

    def _fmt_element(self):
        return f'''{"".join(e._fmt()[0] for e in self._elements)}'''



class BaseQuerySet(AbstractQueryStatement):
    def __init__(self, query_statements=None):
        if query_statements is None:
            self.query_statements = list()
        else:
            self.query_statements = query_statements # gross

    @property
    def _output(self):
        _outputs = []
        for qs in self.query_statements:
            if isinstance(qs._outputs, list):
                _outputs.extend(qs._outputs)
            else:
                _outputs.append(qs._outputs)
        return tuple(sorted(list(set(_outputs)), key=lambda elm: elm._level))

    def __iter__(self):
        return iter(self.query_statements)

    def __repr__(self):
        return f'''({self._fmt_statement()});'''


class UnionQuerySet(BaseQuerySet):
    def _fmt_statement(self):
        return f'''{"".join(repr(qs) for qs in self.query_statements)}'''


class DifferenceQuerySet(BaseQuerySet):
    def __init__(self, query_statements=None):
        assert len(query_statements) == 2
        super().__init__(query_statements)
        self.minued, self.subtrahend = query_statements

    def _fmt_statement(self):
        return f'''{self.minued} - {self.subtrahend}'''





from enum import Enum
from collections import OrderedDict, defaultdict
import functools
import csv

from .base import Node, Way, Relation, GenericElementQuery, CompoundQuery
from .settings import QuerySettings
from .filters import BboxFilter, TagFilter, IdFilter, UserFilter
from .formats import OutputFormatter, opf

import requests


__all__ = ["NodeQuery",
           "WayQuery",
           "RelationQuery",
           "NodeWayQuery",
           "NodeRelationQuery",
           "WayRelationQuery",
           "NodeWayRelationQuery",
           "QueryBuilder",
           "tagstring_to_oql",
           "nominatim_from_csv"]


class NodeQuery(GenericElementQuery):
    _element = Node


class WayQuery(GenericElementQuery):
    _element = Way


class RelationQuery(GenericElementQuery):
    _element = Relation


class NodeWayQuery(CompoundQuery):
    _elements = (Node, Way)


class NodeRelationQuery(CompoundQuery):
    _elements = (Node, Relation)


class WayRelationQuery(CompoundQuery):
    _elements = (Way, Relation)


class NodeWayRelationQuery(CompoundQuery):
    _elements = (Node, Way, Relation)



class QueryRegister:
    def __init__(self, qb):
        self.qb = qb
        self.qc = list()
        self.r = OrderedDict()

    def __getattr__(self, qname):
        try:
            return self.r[qname]
        except KeyError:
            raise NameError("Named query: {} does not exist".format(qname))

    def append(self, qs, name=None):
        if not self.qc:
            assert qs._input is None
        if not name:
            self.qc.append(qs)
        else:
            self.r[name] = qs
            self.qc.append(name)

    @property
    def _dset(self):
        _ = reversed(self.qc)
        while True:
            item = next(_)
            if item not in self.r.keys():
                return item

    @property
    def output(self):
        return self._dset._output

    def __repr__(self):
        return f'''{"".join(repr(qs) for qs in self.qc)}'''



class QueryBuilder:
    overpass_endpoint = 'http://overpass-api.de/api/interpreter'

    def __init__(self, name="default", basic_optimize=True, **kwargs):
        self.name = name
        self.settings = QuerySettings(**kwargs)
        self.output_mode = OutputFormatter()
        self.qsx = QueryRegister(self)
        if basic_optimize:
            self.output_mode.SORTORDER = opf.SortOrder.QUADTILE

    @property
    def GlobalBoundingBox(self):
        return self.settings.bbox

    @GlobalBoundingBox.setter
    def GlobalBoundingBox(self, bbox):
        self.settings.bbox = bbox

    def include_geometries(self):
        self.output_mode.GEOMETRY = opf.Geometry.FULL_GEOM

    def include_boundingboxes(self):
        self.output_mode.GEOMETRY = opf.Geometry.BOUNDING_BOX

    def include_centerpoints(self):
        self.output_mode.GEOMETRY = opf.Geometry.CENTER_POINT

    @property
    def raw_query_string(self):
        return f'''{self.settings}{self.qsx}{self.output_mode}'''

    @classmethod
    def from_query(cls, query, *args, **kwargs):
        inst = cls(*args, **kwargs)
        cls._query = query

    def request(self):
        return requests.get(self.overpass_endpoint, data=self.raw_query_string)


def tagstring_to_oql(tagstring):
    dd = defaultdict(list)
    s = [_.strip().split("=") for _ in tagstring.split("OR")]
    for k, v in s:
        dd[k].append(v)
    return functools.reduce(lambda x, y: x + y, [TagFilter(k,  vals=v) for k, v in dd.items()])

def nominatim_from_csv(path):
    d = dict()
    with open(path) as f:
        reader = csv.DictReader(f)
        for row in reader:
            d[row['word']] = row['query']
    return d






import os
import sys
import json
import math
import datetime
from functools import partial, partialmethod
import types
import dataclasses
from pathlib import Path

import boto3
import botocore
import s3fs

import numpy as np
import networkx as nx
import rasterio
from rasterio.plot import show as rashow
import shapely.ops as ops
import shapely.geometry as geom
from pyproj import CRS, Proj, Transformer

import tiletanic as tt

from .utils import itrreduce, to_gjson

WORLD_CRS = CRS.from_epsg(4326)
MAP_CRS = CRS.from_epsg(3857)

# Following section defines custom Tiletanic TileSchemes that I have to put here
# I guess
class WorldNorthernUTMTiling(tt.tileschemes.UTMTiling):
    """
    Transforms Southern Hemisphere Polar axes coordinates onto
    the Northern Hemisphere Polar axes, which defines the origin
    towards the Southern Pole and the positive upper bound to the North.
    Amounts to a simple translation by the UTM zone height.
    Convenience subclass that modifies the two public methods
    used by the other geographic methods so as to make the change
    at the source where the use case is relevant.
    """

    def ul(self, *tile):
        """Returns the upper left coordinate of the (x, y, z) tile, and
        translates the y coordinate to UTM Northern Hemisphere South-North
        bounds if the tile is defined by UTM Southern Hemisphere bounds.
        Args:
            *tile: (x, y, z) tile coordinates or a Tile object we want
                   the upper left geospatial coordinates of.
        Returns:
            The upper left geospatial coordiantes of the input tile.
        """

        coords = super().ul(*tile)
        if self.quadkey(*tile)[0] in ('2', '3'):
            coords = tt.base.Coords(coords.x, coords.y + 10_000_000)
        return coords

    def br(self, *tile):
        """Returns the bottom right coordinate of the (x, y, z) tile, and
        translates the y coordinate to UTM Northern Hemisphere South-North
        bounds if the tile is defined by UTM Southern Hemisphere bounds.
        Args:
            *tile: (x, y, z) tile coordinates or a Tile object we want
                   the bottom right geospatial coordinates of.
        Returns:
            The bottom right geospatial coordiantes of the input tile.
        """


        coords = super().br(*tile)
        if self.quadkey(*tile)[0] in ('2', '3'):
            coords = tt.base.Coords(coords.x, coords.y + 10_000_000)
        return coords



class WNUTM5kmTiling(tt.tileschemes.UTM5kmTiling, WorldNorthernUTMTiling):
    pass

# End tileschemes extensions


# Bounds are always (lonmin, latmin, lonmax, latmax) from shit
# eg (xmin, ymin, xmax, ymax)

class GeoTileTransformer(object):
    def __init__(self, tile, to_crs):
        self.tile = tile
        self.to_crs = to_crs

    def __enter__(self):
        self.tile.set_crs(self.to_crs)
        return self.tile

    def __exit__(self, *args):
        self.tile.reset_crs()



class WGS84Transformer(GeoTileTransformer):
    def __init__(self, tile):
        GeoTileTransformer.__init__(self, tile, WORLD_CRS)



class GeoTile(object):
    dtype = [('ix', int), ('iy', int), ('zoom', int)]

    def __init__(self, xi, yi, zoom, tiler=None):
        self.xi = xi
        self.yi = yi
        self.zoom = zoom
        self._tiler = tiler
        self._qk = None
        if self.quadkey[0] in ('0', '1'):
            self.__crs__ = tiler._crs_n
            self._tr = tiler._tr_n
        else:
            self.__crs__ = tiler._crs_s
            self._tr = tiler._tr_s
        self._crs = self.__crs__
        self.__gi__ = geom.box(*self.utm_bounds).__geo_interface__
        self._gi = self.__gi__

    @property
    def utm_zone(self):
        return self._tiler.utm_zone

    def parent(self):
        return self._tiler.tile_parent(self.asTile)

    def children(self):
        return self._tiler.tile_children(self.asTile)

    @property
    def quadkey(self):
        if self._qk is None:
            self._qk = self._tiler.quadkey_from_tile(self.asTile)
        return self._qk

    @property
    def asTile(self):
        return tt.base.Tile(self.xi, self.yi, self.zoom)

    @property
    def crs(self):
        return self._crs

    def set_crs(self, to_crs):
        ccrs = self._crs
        self._crs = to_crs
        if ccrs == to_crs:
            return
        if ccrs == self.__crs__ and to_crs == WORLD_CRS:
            tr = self._tr
        else:
            tr = Transformer.from_crs(ccrs, to_crs, always_xy=True).transform
        self._gi = ops.transform(tr, geom.shape(self)).__geo_interface__

    toWGS84 = partialmethod(set_crs, WORLD_CRS)

    def reset_crs(self):
        self._crs = self.__crs__
        self._gi = self.__gi__

    @property
    def width_meters(self):
        return self._tiler.tile_size_at_zoom(self.zoom)

    @property
    def length_meters(self):
        return self.width_meters

    @property
    def utm_bounds(self):
        return self._tiler._tiler.bbox(self.asTile)

    @property
    def bounds(self):
        return geom.shape(self).bounds

    @property
    def __geo_interface__(self):
        return self._gi

    @classmethod
    def _from_tile(cls, tile, tiler=None):
        return cls(tile.x, tile.y, tile.z, tiler=tiler)

    @property
    def asShape(self):
        return geom.asShape(self)

    @property
    def z(self):
        return self.zoom

    def __iter__(self):
        return iter([self.xi, self.yi, self.zoom])





class ProjectedUTMTiling(object):

    def __init__(self, zone=None, tiler=None):
        self.utm_zone = zone
        self._tiler = tiler
        self._tiling_zoom_level = tiler.zoom
        self._tiling_physical_length = tiler.tile_size
        self._tf = partial(GeoTile._from_tile, tiler=self)
        self._crs_n = CRS.from_epsg(f'326{self.utm_zone:02d}')
        self._crs_s = CRS.from_epsg(f'327{self.utm_zone:02d}')
        self._tr_n = Transformer.from_crs(self._crs_n, WORLD_CRS, always_xy=True).transform
        self._tr_s = Transformer.from_crs(self._crs_s, WORLD_CRS, always_xy=True).transform

    def tile_from_xy(self, xcoord, ycoord, zoom=12):
        tile = self._tiler.tile(xcoord, ycoord, zoom)
        return self._tf(tile)

    def tile_from_quadkey(self, qk):
        tile = self._tiler.quadkey_to_tile(qk)
        return self._tf(tile)

    def quadkey_from_tile(self, *tile):
        return self._tiler.quadkey(*tile)

    def tile_parent(self, *tile):
        tile = self._tiler.parent(*tile)
        return self._tf(tile)

    def tile_children(self, *tile):
        children = self._tiler.children(*tile)
        return [self._tf(tile) for tile in children]

    def tile_length_at_zoom(self, zoom_level):
        return self._tiling_physical_length / 2**(zoom_level - self._tiling_zoom_level)

    def zoom_from_tile_dim(self, dimension):
        pass

    def __getattr__(self, atr):
        return getattr(self._tiler, atr)

    def dims_from_zoom(self, zoom, init_dim=16384, init_zoom=12):
        return init_dim / 2**(zoom - init_zoom)



# Accessor api


class CanvasClient(object):
    def __init__(self, bucket, local=True, s3conn=None):
        self.bucket = bucket
        self.local = local
        if not local:
            if s3conn is None:
                s3conn = s3fs.S3FileSystem(anon=False)
        self.s3conn = s3conn

    def list_dir(self, *args, **kwargs):
        if self.local:
            return os.listdir(*args, **kwargs)
        return self.s3conn.ls(*args, **kwargs)



class TileCollection(CanvasClient):

    def __init__(self, utm_zone, tiler=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.utm_zone = utm_zone
        self.tiler = ProjectedUTMTiling(zone=utm_zone, tiler=tiler)
        self.cvg = nx.Graph(store=self.bucket, utm_zone=utm_zone)
        self.qk_path = os.path.join(self.bucket, str(utm_zone))
        self._tile_quadkeys_fetched = False
        self._tile_cogs_fetched = False
        self._qksample = self.tile_quadkeys[0]
        self._gtm = None
        self.__gi__ = None

    @property
    def tile_quadkeys(self):
        if not self._tile_quadkeys_fetched:
            qk_paths = self.list_dir(self.qk_path)
            for qkp in qk_paths:
                qk = Path(qkp).parts[-1]
                self.cvg.add_node(qk, obj="quadkey")
            self._tile_quadkeys_fetched = True
        return [n for n in self.cvg.nodes() if self.cvg.nodes[n]['obj'] == "quadkey"]

    @property
    def tile_cogs(self):
        if not self._tile_cogs_fetched:
            for qk in self.tile_quadkeys:
                cog_path = os.path.join(self.qk_path, qk)
                cog_files = self.list_dir(cog_path)
                for fp in cog_files:
                    catid = Path(fp).stem.split("-")[0]
                    self.cvg.add_node(catid, obj="catalog_id")
                    self.cvg.add_edge(catid, qk)
            self._tile_cogs_fetched = True
        return [n for n in self.cvg.nodes() if self.cvg.nodes[n]['obj'] == 'catalog_id']

    @property
    def nqks(self):
        return len(self.tile_quadkeys)

    @property
    def ncatids(self):
        return len(self.tile_cogs)

    @property
    def area_coverage(self, units='m'):
        return str((5_000 * 5_000) * self.nqks) + ' ' + units + '^2'

    def __getitem__(self, quadkey):
        if quadkey not in self.tile_quadkeys:
            raise KeyError
        return self.tiler.tile_from_quadkey(quadkey)

    def __setitem__(self, item):
        raise NotImplementedError

    def __iter__(self):
        tiles = np.array([self.tiler._tiler.quadkey_to_tile(qk)
                          for qk in self.tile_quadkeys], dtype=GeoTile.dtype)
        for tile in np.sort(tiles, order=['iy', 'ix']):
            yield self.tiler._tf(tt.base.Tile(*tile))

    def iter_geoms(self):
        for tile in iter(self):
            with WGS84Transformer(tile) as geotile:
                yield geotile.asShape

    @property
    def __geo_interface__(self):
        if self.__gi__ is None:
            self.__gi__ = itrreduce(ops.unary_union,
                                    self.iter_geoms()).__geo_interface__
        return self.__gi__

    @property
    def asShape(self):
        return geom.asShape(self)

    @property
    def centroid(self):
        return self.asShape.centroid



class CanvasCollection(CanvasClient):
    def __init__(self, aoi=None, tilescheme=WNUTM5kmTiling, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aoi = aoi
        self.tiler = tilescheme()
        self._zlut = dict()
        self._zone_paths_fetched = False

    def canvas_zones(self):
        if not self._zone_paths_fetched:
            self._zone_paths = self.list_dir(self.bucket)
            self._zone_paths_fetched = True
            self.zones = sorted([int(Path(zp).parts[-1]) for zp in self._zone_paths if Path(zp).parts[-1] not in ('models', 'output')])
        return self.zones

    def zone(self, zone, *args, **kwargs):
        if zone not in self.canvas_zones():
            raise OSError("Zone {} not in path".format(zone))
        if zone in self._zlut: return self._zlut[zone]
        self._zlut[zone] = TileCollection(zone, bucket=self.bucket, local=self.local, s3conn=self.s3conn, tiler=self.tiler, **kwargs)
        return self._zlut[zone]

    def descriptions(self):
        for z in self.canvas_zones():
            print(z)
            print("Total data area covered: " + self.zone(z).area_coverage)
            print("Total number of quadkey zones: " + str(self.zone(z).nqks))
            print("Total number of catalog ids tiled: " + str(self.zone(z).ncatids))
            print("\n")




from rasterio.windows import Window
import shapely.geometry as geom
import rasterio
import csv


def balh(impath):
    with rasterio.open(impath) as src:
        plt.figure(figsize=(10, 10))
        arr = src.read(window=Window(512, 512, 512*2, 512*2))
        rashow(arr)


def itrreduce(fn, iterable, initializer=None):
    it = iter(iterable)
    if initializer is None:
        value = next(it)
    else:
        value = initializer
    for element in it:
        value = fn([value, element])
    return value


def to_gjson(tile):
    geodata = geom.mapping(tile)
    gj = {"type": "Feature",
         "properties": {"tile_id": "+".join([str(tile.utm_zone), tile.quadkey]),
                        "ix": tile.xi,
                        "iy": tile.yi,
                        "zoom": tile.zoom},
         "geometry": geodata}
    return gj





from enum import Enum


class OutputFormats:
    class Verbosity(Enum):
        GENERIC = "body"
        CONCISE = "skel"
        BRIEF = "ids"
        VERBOSE = "meta"

    class Geometry(Enum):
        FULL_GEOM = "geom"
        BOUNDING_BOX = "bb"
        CENTER_POINT = "center"

    class SortOrder(Enum):
        OBJECT_ID = "asc"
        QUADTILE = "qt"

    # Can set defaults and stuff here

opf = OutputFormats()

class OutputFormatter:
    def __init__(self,
                 verbosity=opf.Verbosity.GENERIC,
                 geometry=None,
                 sortorder=opf.SortOrder.OBJECT_ID):

        self._v = verbosity
        self._g = geometry
        self._s = sortorder

    @property
    def VERBOSITY(self):
        return self._v

    @VERBOSITY.setter
    def VERBOSITY(self, mode):
        self._v = mode

    @property
    def GEOMETRY(self):
        return self._g

    @GEOMETRY.setter
    def GEOMETRY(self, mode):
        self._g = mode

    @property
    def SORTORDER(self):
        return self._s

    @SORTORDER.setter
    def SORTORDER(self, mode):
        self._s = mode

    def __repr__(self):
        s = "out"
        modes = []
        if self._v != opf.Verbosity.GENERIC:
            modes.append(self._v.value)
        if self._g is not None:
            modes.append(self._g.value)
        if self._s == opf.SortOrder.QUADTILE:
            modes.append(self._s.value)
        if modes:
            s = f'''{s} {" ".join(m for m in modes)}'''
        return s + ";"




import os
import re
import json
import csv
from tempfile import NamedTemporaryFile
from collections import defaultdict
import requests
import pandas as pd
from bs4 import BeautifulSoup

ti_stat_ep = "https://taginfo.openstreetmap.org/api/4/tag/stats"
ti_list_ep = "https://taginfo.openstreetmap.org/api/4/tags/list"
OSM_FEATURES_URL = "https://wiki.openstreetmap.org/wiki/Map_Features#Primary_features"

def write_url(url, outfile):
    r = requests.get(url)
    with open(outfile, 'w') as f:
        f.write(r.text)


def _tagname_from_header(tag):
    hdr = tag.find_previous("h3")
    return hdr.find_next("span", class_="mw-headline").attrs["id"]


def _itrtable(t):
    for row in t.find_all('tr'):
        yield row


def _parse_tagstr(s, tagname, td):
    for v in s.split(","):
        try:
            t, v = v.split("=")
            if t != tagname.lower():
                continue
        except ValueError:
            if t != tagname.lower():
                continue
        if v != "": td[tagname][t].append([v, "NA"])
    return td


def _rendered_strikethrough(td):
    return td.find('s') is not None


def _parse_table(tbl, tagname, tbld, icols=(0, 1, 3), filter_strikethrough=True):
    for i, row in enumerate(_itrtable(tbl)):
        if i == 0: # Header, we define ours by index
            writing = True
            continue
        tds = row.findChildren('td')
        if len(tds) == 0: # We deal here with filtering unwanted subheadings like attrs
            subhead = row.findChild('span', class_="mw-headline")
            if subhead: # Sometimes it's not a span like in highways
                writing = "attribute" not in subhead['id'].lower()
                if tagname == "Railway" and "Additional_features" == subhead['id']:
                    writing = False
        else:
            if writing:
                if filter_strikethrough:
                    # test either first two elms, have to test both cause osm
                    # wiki is a bag of farts
                    if _rendered_strikethrough(tds[0]) or _rendered_strikethrough(tds[1]):
                        continue
                key, val, desc = (tds[icol].get_text().strip("\n").strip() for icol in icols)
                # Don't write about user defined crap
                if "user" not in val.lower():
                    tbld[tagname][key].append([val, desc])
    return tbld



def scrape_primary_features(url=OSM_FEATURES_URL, use_cache=False):
    r = requests.get(url)
    soup = BeautifulSoup(r.text, 'html.parser')
    # Get the tables and taglists _above_ the 'Additional Properties' items
    # that tail the data in the doc
    adp = soup.find_all("span", class_="mw-headline", id="Additional_properties").pop()
    # There are two kinds of tables in the doc; taglists which call external
    # javascript and can be identified by their div class, and html tables which
    # can be parsed via row elements. Statistical data much be called manually
    # here using the osm api after the key, value pairs are collected.

    # Get the tags
    tagd = defaultdict(lambda: defaultdict(list))
    tags = adp.find_all_previous('div', class_='taglist')

    tbld = defaultdict(lambda: defaultdict(list))
    tbls = adp.find_all_previous('table', class_='wikitable')
    # To get the correct tagname, look for the span under the previous <h3>
    # element. This groups tags correctly instead of under descriptive
    # categories, eg Places instead of Other Places etc
    for tag in tags:
        tagname = _tagname_from_header(tag)
        datastr = tag.attrs['data-taginfo-taglist-tags']
        tagd = _parse_tagstr(datastr, tagname, tagd)

    # We can get the descriptions and other things from the tables
    for tbl in tbls:
        tagname = _tagname_from_header(tbl)
        tbld = _parse_table(tbl, tagname, tbld)

    return (tagd, tbld)





import typing


class Filter:
    def _fmt(self):
        raise NotImplementedError

    def __repr__(self):
        return self._fmt()

    def __add__(self, other):
        if isinstance(other, str):
            other = UserFilter(other)
        if isinstance(other, (Filter, CompoundFilter, typing.Sequence)):
            return CompoundFilter([self, other])
        raise NotImplementedError

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        return self.__add__(other)


class CompoundFilter:
    def __init__(self, filters=None):
        self._filters = list()
        if filters is not None:
            self.filters = filters

    @property
    def filters(self):
        return self._filters

    @filters.setter
    def filters(self, filters):
        if not self._filters:
            self.add_filter(filters)
        else:
            raise AttributeError("Delete existing filters before assignment")

    @filters.deleter
    def filters(self):
        self._filters = list()

    def __iter__(self):
        return iter(self.filters)

    def __repr__(self):
        return f'''{"".join(f._fmt() for f in self.filters)}'''

    def add_filter(self, other):
        if isinstance(other, str):
            other = UserFilter(other)
        if isinstance(other, Filter):
            self._filters.append(other)
            return
        if isinstance(other, CompoundFilter):
            self._filters.extend(other.filters)
            return
        if isinstance(other, typing.Sequence):
            for f in other:
                self.add_filter(f)
        else:
            raise NotImplementedError

    def __add__(self, other):
        if isinstance(other, str):
            other = UserFilter(other)
        if isinstance(other, Filter):
            filters = [f for f in self.filters]
            filters.append(other)
            return CompoundFilter(filters)
        if isinstance(other, CompoundFilter):
            filters = list(itertools.chain(self.filters, other.filters))
            return CompoundFilter(filters)
        else:
            raise NotImplementedError

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        self.add_filter(other)
        return self


class GenericFilter(Filter):
    pass


class UserFilter(GenericFilter):
    def __init__(self, filterstring):
        self._fs = filterstring

    def _fmt(self):
        return self._fs


class IdFilter(GenericFilter):
    def __init__(self, osm_id):
        self._id = osm_id

    @property
    def osm_id(self):
        return self._id

    @osm_id.setter
    def osm_id(self, val):
        self._id = val

    def _fmt(self):
        return f'''(id:{self.osm_id})'''


class BboxFilter(GenericFilter):
    def __init__(self, bbox):
        self.bbox = bbox

    def _fmt(self):
        s, w, n, e = self.bbox
        return f'''({s:.8f},{w:.8f},{n:.8f},{e:.8f})'''


class KeySpec:
    def __init__(self, key, exists=True):
        if key.startswith("!"):
            if not exists:
                raise AttributeError("Input existence conflict")
            key = key.strip("!")
            exists = False
        self.key = key
        self.exists = exists

    def __bool__(self):
        return self.exists

    def __repr__(self):
        s = f'''{self.key}'''
        if not self.exists:
            return "!" + s
        return s



def format_values(vals):
    if isinstance(vals, str):
        return f'''={vals}'''
    if isinstance(vals, list):
        if len(vals) == 1:
            val = vals[0]
            return f'''={val}'''
        else:
            svals = "|".join(vals)
            return f'''~"^({svals})$"'''


class TagFilter(GenericFilter):
    def __init__(self, key, vals=[], exists=True):
        if vals == ['*']:
            vals = None
        if not vals:
            self._key = KeySpec(key, exists=exists)
        else:
            self._key = KeySpec(key)
        self._vals = vals
        self.exists = exists

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, val):
        self._key = KeySpec(key)

    @property
    def values(self):
        return self._vals

    @values.setter
    def values(self, vals):
        self._vals = vals

    def _fmt(self):
        if not self.key.exists:
            return f'''[{self.key}]'''
        if not self.values:
            return f'''[{self.key}]'''
        svals = format_values(self.values)
        if self.exists:
            return f'''[{self.key}{svals}]'''
        return f'''[{self.key}][{self.key}!{svals}]'''







import sys
import os
from setuptools import setup, find_packages

req_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "requirements.txt")
with open(req_path) as f:
    reqs = f.read().splitlines()


setup(name="maxar-skyway",
      version="0.0.1",
      author="Jamison Polackwich",
      author_email='jamie.polackwich@digitalglobe.com',
      description="Python API for making OSM queries",
      url="https://github.com/maxar-analytics/skyway",
      license="MIT",
      packages=find_packages(exclude=['docs', 'tests']),
      install_requires=reqs,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License"
    ],
      python_requires='>3.7',
      )

